///////////////////////////////////////////////////////////////////////
// AC-3 Plugin for VirtualDub 1.8.8+
// Copyright (C) 2008-2011 fccHandler
// Copyright (C) 1998-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

#include <windows.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"

#include "WinFile.h"
#include "AC3Info.h"
#include "InputFileAC3.h"
#include "AudioSourceAC3.h"


AudioSourceAC3::AudioSourceAC3(InputFileAC3 *pp)
	: parentPtr     (pp)
	, mLength       (0)
	, mDuration     (0.0)
	, fIsVBR        (false)
{
	parentPtr->AddRef();

	if (parentPtr->framelist != NULL && parentPtr->frames > 0)
	{
		unsigned long dwTotalBytes;
		int iMaxSample;

		int srate, brate, chans;
		unsigned long i;

		mLength = parentPtr->frames;

		// Calculate total bytes and maximum sample size

		iMaxSample = ac3_sync_info(
			parentPtr->FirstHeader(),
			&srate, &brate, &chans
		);
		dwTotalBytes = iMaxSample;

		for (i = 1; i < mLength; ++i)
		{
			int frmsize = (int)parentPtr->framelist[i].size;
			if (frmsize != iMaxSample)
			{
				fIsVBR = true;
				if (frmsize > iMaxSample) iMaxSample = frmsize;
			}
			dwTotalBytes += frmsize;
		}

		// Total duration in seconds
		mDuration = ((double)mLength * 1536.0) / (double)srate;

		mRawFormat.wFormatTag       = 0x2000;
		mRawFormat.nChannels        = chans;
		mRawFormat.nSamplesPerSec   = srate;
		mRawFormat.nAvgBytesPerSec  = (DWORD)((double)dwTotalBytes / mDuration + 0.5);
		mRawFormat.nBlockAlign      = (WORD)iMaxSample;
		mRawFormat.wBitsPerSample   = 0;
		mRawFormat.cbSize           = 0;
	}
}


AudioSourceAC3::~AudioSourceAC3()
{
	parentPtr->Release();
}


int VDXAPIENTRY AudioSourceAC3::AddRef()
{
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int VDXAPIENTRY AudioSourceAC3::Release()
{
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VDXAPIENTRY AudioSourceAC3::AsInterface(uint32 iid)
{
	if (iid == IVDXAudioSource::kIID)
		return static_cast<IVDXAudioSource *>(this);

	if (iid == IVDXStreamSourceV3::kIID)
		return static_cast<IVDXStreamSourceV3 *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


void VDXAPIENTRY AudioSourceAC3::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo)
{
	if (mLength != 0)
	{
		srcInfo.mSampleRate.mNumerator   = mRawFormat.nSamplesPerSec;
		srcInfo.mSampleRate.mDenominator = 1536;
		srcInfo.mSampleCount             = mLength;
	}
	else
	{
		srcInfo.mSampleRate.mNumerator   = 1;
		srcInfo.mSampleRate.mDenominator = 1;
		srcInfo.mSampleCount             = 0;
	}

	srcInfo.mPixelAspectRatio.mNumerator   = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}


void VDXAPIENTRY AudioSourceAC3::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& info)
{
	GetStreamSourceInfo(info.mInfo);

	if (fIsVBR)
		info.mFlags  = VDXStreamSourceInfoV3::kFlagVariableSizeSamples;
	else
		info.mFlags  = 0;

	info.mfccHandler = 0;
}


bool VDXAPIENTRY AudioSourceAC3::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	unsigned long dwSamples = 0;
	unsigned long dwBytes   = 0;
	unsigned long dwStart   = (unsigned long)lStart64;

	// Is Start out of range?
	if (dwStart >= mLength || lCount == 0)
	{
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead)   *lBytesRead   = 0;
		return true;
	}
	
	// Is output buffer too small?
	if (cbBuffer < mRawFormat.nBlockAlign)
	{
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead)   *lBytesRead   = 0;
		return false;
	}

	// Buffer size query?
	if (lpBuffer == NULL)
	{
		while (lCount-- > 0)
		{
			dwBytes += (int)parentPtr->framelist[dwStart].size;
			++dwSamples;
			if (++dwStart >= mLength) break;
		}

		if (dwBytes < mRawFormat.nBlockAlign)
			dwBytes = mRawFormat.nBlockAlign;

		if (lSamplesRead) *lSamplesRead = dwSamples;
		if (lBytesRead)   *lBytesRead   = dwBytes;
		return true;
	}

	while (lCount > 0)
	{
		int frmsize = (int)parentPtr->framelist[lStart64].size;
		if ((uint32)frmsize > cbBuffer) break;

		frmsize = parentPtr->ReadSample(lpBuffer, (unsigned long)lStart64);
		if (frmsize <= 0) break;

		dwBytes += frmsize;
		++dwSamples;
		++lStart64;
		if (lStart64 >= (sint64)mLength) break;

		--lCount;
		lpBuffer = (char *)lpBuffer + frmsize;
		cbBuffer -= frmsize;
	}

	if (lSamplesRead) *lSamplesRead = dwSamples;
	if (lBytesRead)   *lBytesRead   = dwBytes;
	return true;
}


const void *VDXAPIENTRY AudioSourceAC3::GetDirectFormat()
{
	if (mLength != 0) return &mRawFormat;
	return NULL;
}


int VDXAPIENTRY AudioSourceAC3::GetDirectFormatLen()
{
	if (mLength != 0) return sizeof(mRawFormat);
	return 0;
}


IVDXStreamSource::ErrorMode VDXAPIENTRY AudioSourceAC3::GetDecodeErrorMode()
{
	return IVDXStreamSource::kErrorModeReportAll;
}


void VDXAPIENTRY AudioSourceAC3::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
}


bool VDXAPIENTRY AudioSourceAC3::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	return mode == IVDXStreamSource::kErrorModeReportAll;
}


bool VDXAPIENTRY AudioSourceAC3::IsVBR()
{
	return false;
}


sint64 VDXAPIENTRY AudioSourceAC3::TimeToPositionVBR(sint64 us)
{
	if (mLength != 0)
	{
		return (sint64)(
			((double)us * (double)mRawFormat.nSamplesPerSec)
			/ 1536000000.0 + 0.5
		);
	}
	return 0;
}


sint64 VDXAPIENTRY AudioSourceAC3::PositionToTimeVBR(sint64 samples)
{
	if (mLength != 0)
	{
		return (sint64)(
			( (double)samples * 1536000000.0)
			/ (double)mRawFormat.nSamplesPerSec + 0.5
		);
	}
	return 0;
}


void VDXAPIENTRY AudioSourceAC3::GetAudioSourceInfo(VDXAudioSourceInfo& info)
{
	info.mFlags = 0;
}

