///////////////////////////////////////////////////////////////////////
// AC-3 Plugin for VirtualDub 1.8.8+
// Copyright (C) 2008-2011 fccHandler
// Copyright (C) 1998-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////


/********************************************************************
 * Windows file I/O functions, wrapped in a class.
 * Files larger than 4GB are supported on NTFS.
 ********************************************************************/


#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "WinFile.h"


//#define INPAGE(x)		(x & (~(WF_BUFFERSIZE-1)))
//#define INOFFSET(x)		(x & (WF_BUFFERSIZE-1))


LPSTR WinFile::WcharToAnsi(LPCWSTR lpszWchar) const
{
	LPSTR ret = NULL;

	if (lpszWchar != NULL)
	{
		int len = WideCharToMultiByte(CP_ACP, 0, lpszWchar, -1, NULL, 0, NULL, NULL);
		if (len > 0)
		{
			ret = new char[len];
			if (ret != NULL)
			{
				if (WideCharToMultiByte(CP_ACP, 0, lpszWchar, len, ret, len, NULL, NULL) != len)
				{
					delete[] ret;
					ret = NULL;
				}
			}
		}
	}

	return ret;
}


WinFile::WinFile()
{
	mHandle = INVALID_HANDLE_VALUE;

	mPos.QuadPart  = 0;
	mSize.QuadPart = 0;
}


WinFile::~WinFile()
{
	if (mHandle != INVALID_HANDLE_VALUE) CloseHandle(mHandle);
}


/********************************************************************
 *
 *	How Windows handles some aberrant situations:
 *
 *	Setting the file pointer to absolute -10 returns 0xFFFFFFFF
 *	and error code 87 (ERROR_INVALID_PARAMETER), and does
 *	not move the file pointer (apparently)
 *
 *	Setting the file pointer to relative -10 returns 0xFFFFFFFF
 *	and error code 131 (ERROR_NEGATIVE_SEEK), and does
 *	not move the file pointer
 *
 *	Setting the file pointer AT or BEYOND the end of file succeeds.
 *
 *	Attempting to read past the end of file does not cause
 *	an error, but reads zero bytes.
 *
 ********************************************************************/

void WinFile::Refresh()
{
	DWORD dwRead = mPos.HighPart;
	DWORD dwLo   = mPos.LowPart & (~(WF_BUFFERSIZE - 1));

	SetFilePointer(mHandle, dwLo, (PLONG)&dwRead, FILE_BEGIN);

	if (ReadFile(mHandle, mBuffer, WF_BUFFERSIZE, &dwRead, NULL) == FALSE)
	{
		dwRead = 0;
	}

	if (dwRead < WF_BUFFERSIZE)
	{
		memset(mBuffer + dwRead, 0, WF_BUFFERSIZE - dwRead);
	}
}


HANDLE WinFile::DualCreateFile(
	LPCWSTR lpFileName,
	DWORD dwDesiredAccess,
	DWORD dwShareMode,
	LPSECURITY_ATTRIBUTES lpSecurityAttributes,
	DWORD dwCreationDisposition,
	DWORD dwFlagsAndAttributes,
	HANDLE hTemplateFile) const
{
	HANDLE h;

	if (GetVersion() < 0x80000000U)
	{
		h = CreateFileW(lpFileName, dwDesiredAccess,
			dwShareMode, lpSecurityAttributes,
			dwCreationDisposition, dwFlagsAndAttributes,
			hTemplateFile);
	}
	else
	{
		LPSTR tmp = WcharToAnsi(lpFileName);

		h = INVALID_HANDLE_VALUE;

		if (tmp != NULL)
		{
			h = CreateFileA(tmp, dwDesiredAccess, dwShareMode,
				lpSecurityAttributes, dwCreationDisposition,
				dwFlagsAndAttributes, hTemplateFile);

			delete[] tmp;
		}
	}

	return h;
}


bool WinFile::Open(const wchar_t *lpszFileName)
{
	DWORD   dwLo;
	DWORD   dwHi;
	HANDLE  h;
	bool    ret;

	h       = INVALID_HANDLE_VALUE;
	ret     = false;

	if (lpszFileName  == NULL ) goto Abort;
	if (*lpszFileName == L'\0') goto Abort;

	h = DualCreateFile(
		(LPCWSTR)lpszFileName,
		GENERIC_READ,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL
	);

	if (h == INVALID_HANDLE_VALUE) goto Abort;

	dwLo = GetFileSize(h, &dwHi);
	if (dwLo == INVALID_FILE_SIZE)
	{
		if (GetLastError() != NO_ERROR) goto Abort;
	}
	if (dwHi & 0x80000000) goto Abort;

	if (mHandle != INVALID_HANDLE_VALUE) CloseHandle(mHandle);
	mHandle = h;
	h = INVALID_HANDLE_VALUE;

	mPos.QuadPart  = 0;
	mSize.LowPart  = dwLo;
	mSize.HighPart = dwHi;

	Refresh();
	ret = true;

Abort:
	if (h != INVALID_HANDLE_VALUE) CloseHandle(h);
    return ret;
}


bool WinFile::Seek(__int64 iAbsPos)
{
	bool ret = false;

	if (mHandle != INVALID_HANDLE_VALUE && iAbsPos >= 0)
	{
		if (iAbsPos > mSize.QuadPart) iAbsPos = mSize.QuadPart;

		ret = (
			(iAbsPos       & (~(WF_BUFFERSIZE - 1))) !=
			(mPos.QuadPart & (~(WF_BUFFERSIZE - 1)))
		);

		mPos.QuadPart = iAbsPos;

		if (ret) Refresh();
		ret = true;
	}

	return ret;
}


int WinFile::Read(void *lpBuffer, int iBytes)
{
	int remains, offs;
	int ret = 0;

	if (mHandle == INVALID_HANDLE_VALUE) goto Abort;
	if (lpBuffer == NULL) goto Abort;
	if (iBytes <= 0) goto Abort;

	// Don't read past EOF
	{
		__int64 remains64 = mSize.QuadPart - mPos.QuadPart;
		if (remains64 <= 0) goto Abort;
		if ((__int64)iBytes > remains64) iBytes = (int)remains64;
	}

	offs    = mPos.LowPart & (WF_BUFFERSIZE - 1);
	remains = WF_BUFFERSIZE - offs;

	while (true)
	{
		if (remains > iBytes) remains = iBytes;
		memcpy(lpBuffer, mBuffer + offs, remains);

		ret += remains;
		mPos.QuadPart += remains;
		iBytes -= remains;
		if (iBytes <= 0) break;

		lpBuffer = (char *)lpBuffer + remains;
		Refresh();
		offs = 0;
		remains = WF_BUFFERSIZE;
	}

	if ((mPos.LowPart & (WF_BUFFERSIZE - 1)) == 0) Refresh();

Abort:
	return ret;
}


int WinFile::ReadByte()
{
	int ret = 0;

	if (mHandle != INVALID_HANDLE_VALUE)
	{
		if (mPos.QuadPart < mSize.QuadPart)
		{
			int offs = mPos.LowPart & (WF_BUFFERSIZE - 1);
			ret = mBuffer[offs];
			++mPos.QuadPart;
			if (offs == (WF_BUFFERSIZE - 1)) Refresh();
		}
	}

	return ret;
}
